
#define GLES20EMU_CPP
#include "GLES20Emu.h"

#if defined(_WIN32) || defined (_WIN64)
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <GL/wglext.h> // wglGetProcAddress()
#else // _WIN32, _WIN64
#include <GL/glx.h>
//#include <GL/glxext.h> // my glxext.h doesn't cope nicely with some unices; gotta get another version

#ifndef GLX_VERSION_1_4

/**
 *  @def glXGetProcAddress
 *  @brief get proc address on linux platform
 *  @param[in] x is name of the function to retrieve
 */
#define glXGetProcAddress(x) (*glXGetProcAddressARB)((const GLubyte*)x)

#endif // !GLX_VERSION_1_4

/**
 *  @def wglGetProcAddress
 *  @brief windows get proc address lookalike on linux platform
 *  @param[in] x is name of the function to retrieve
 */
#define wglGetProcAddress(x) (*glXGetProcAddress)((GLubyte*)(x))
#endif // _WIN32, _WIN64

#ifdef NULL_FUNCTION_POINTERS
#define INIT_FUNC_POINTERS = 0
#else
#define INIT_FUNC_POINTERS 
#endif

namespace stl_ut {

template <class CContainer>
inline bool Reserve_N(CContainer &r_vec, size_t n) throw()
{
    try {
        r_vec.reserve(n);
        return r_vec.capacity() >= n;
        // reserve and check
    } catch(std::bad_alloc&) {
        // not enough memory
        return false;
    }
}

template <class CContainer>
inline bool Resize_To_N(CContainer &r_vec, size_t n) throw()
{
    try {
        r_vec.resize(n);
        return true;
        // just resize
    } catch(std::bad_alloc&) {
        // not enough memory
        return false;
    }
}

template <class _String, class _TChar>
bool AssignCStr(_String &r_s_dest, _TChar *p_s_src) throw()
{
    if(!Reserve_N(r_s_dest, strlen(p_s_src)))
        return false;
    // make sure there's space in the string

    r_s_dest.assign(p_s_src);
    // append

    return true;
}

bool Format(std::string &r_s_result, const char *p_s_fmt, ...)
{
    va_list ap;

    size_t n_size = strlen(p_s_fmt) + 1;
    if(!Resize_To_N(r_s_result, n_size))
        return false;
    // alloc str as long as formatting str (should be enough in some cases)

    for(;;) {
        va_start(ap, p_s_fmt);
#if defined(_MSC_VER) && !defined(__MWERKS__) // msvc
#if _MSC_VER >= 1400
        int n = _vsnprintf_s(&r_s_result[0], n_size, _TRUNCATE, p_s_fmt, ap);
#else //_MSC_VER >= 1400
        int n = _vsnprintf(&r_s_result[0], n_size - 1, p_s_fmt, ap);
#endif //_MSC_VER >= 1400
        // maximum characters to write, not the size of buffer
#else
        int n = vsnprintf(&r_s_result[0], n_size, p_s_fmt, ap);
#endif
        va_end(ap);
        // try to sprintf

        if(n >= 0 && unsigned(n) < n_size) {
#if defined(_MSC_VER) && !defined(__MWERKS__) // msvc
            Resize_To_N(r_s_result, n);
#else
            Resize_To_N(r_s_result, strlen(r_s_result.data())); // doesn't need c_str(), terminating zero is already there
#endif
            assert(r_s_result.length() == strlen(r_s_result.data()));
            return true;
        }
        // see if we made it

        if(n >= 0) // glibc 2.1
            n_size = n + 1; // precisely what is needed
        else // glibc 2.0, msvc
            n_size *= 2;

        if(!Resize_To_N(r_s_result, n_size))
            return false;
        // realloc to try again
    }
}

}

/*
 *                              === CGLES20ExtensionHandler ===
 */

bool CGLES20ExtensionHandler::b_Support_OpenGL(int n_min_major, int n_min_minor)
{
    std::pair<int, int> t_version = t_OpenGL_Version();
    int n_major = t_version.first,
        n_minor = t_version.second;
    if(n_major > n_min_major)
        return true;
    if(n_major == n_min_major && n_minor >= n_min_minor)
        return true;
    return false;
}

int CGLES20ExtensionHandler::n_OpenGL_Version()
{
    std::pair<int, int> t_version = t_OpenGL_Version();
    return t_version.first * 10 + t_version.second;
}

std::pair<int, int> CGLES20ExtensionHandler::t_OpenGL_Version()
{
    int n_major, n_minor;
    const char *p_s_version;
    p_s_version = (const char*)glGetString(GL_VERSION);
    if(!p_s_version) {
#if defined(_WIN32) || defined (_WIN64)
        MessageBoxA(0, "OpenGL extension string is 0!", "CGLES20ExtensionHandler", MB_OK);
#endif // _WIN32 || _WIN64
        return std::make_pair(-1, -1);
    }

#if defined(_MSC_VER) && !defined(__MWERKS__) && _MSC_VER >= 1400
    if(sscanf_s(p_s_version, "%d.%d", &n_major, &n_minor) != 2) {
#else //_MSC_VER && !__MWERKS__ && _MSC_VER >= 1400
    if(sscanf(p_s_version, "%d.%d", &n_major, &n_minor) != 2) {
#endif //_MSC_VER && !__MWERKS__ && _MSC_VER >= 1400
#if defined(_WIN32) || defined (_WIN64)
        MessageBoxA(0, "OpenGL version string malformed!", "CGLES20ExtensionHandler", MB_OK);
#endif // _WIN32 || _WIN64
        //_ASSERTE(0); // error - OpenGL version string malformed!
        return std::make_pair(-1, -1);
    }

    return std::make_pair(n_major, n_minor);
}

int CGLES20ExtensionHandler::n_OpenGL_Version_Major()
{
    return t_OpenGL_Version().first;
}

int CGLES20ExtensionHandler::n_OpenGL_Version_Minor()
{
    return t_OpenGL_Version().second;
}

static const GLubyte* (GLApi *glGetStringi_ptr)(GLenum name, GLuint index) = 0;
#define GL_NUM_EXTENSIONS	0x821D /**< @brief OpenGL 3.0 enum */
// that's how it's done on new hw

bool CGLES20ExtensionHandler::b_SupportedExtension(const char *p_s_extension_name)
{
	if(glGetStringi_ptr || (glGetStringi_ptr =
	   (const GLubyte*(GLApi*)(GLenum, GLuint))wglGetProcAddress("glGetStringi"))) {
		// try to fetch glGetStringi()

		int n_extension_num;
		glGetIntegerv(GL_NUM_EXTENSIONS, &n_extension_num);
		for(int i = 0; i < n_extension_num; ++ i) {
			const char *p_s_ext_name = (const char*)glGetStringi_ptr(GL_EXTENSIONS, i);
			if(!p_s_ext_name)
				continue; // ?

			if(!strcmp(p_s_ext_name, p_s_extension_name))
				return true; // ... and there it is!
		}
		// this tends to be slightly faster
	} else {
		const char *p_s_extensions;
		if(!(p_s_extensions = (const char*)glGetString(GL_EXTENSIONS)))
			return false;
		const char *p_s_search_str = p_s_extensions;
		do {
			if((p_s_search_str = strstr(p_s_search_str, p_s_extension_name))) {
				if(p_s_search_str > p_s_extensions && !isspace(*(p_s_search_str - 1)))
					continue;
				if(strlen(p_s_search_str + strlen(p_s_extension_name)) &&
				   !isspace(*(p_s_search_str + strlen(p_s_extension_name))))
					continue;

				return true;
			}
		} while(p_s_search_str && *(++ p_s_search_str));
	}

	return false;
}

bool CGLES20ExtensionHandler::Get_ExtensionString(std::string &r_s_extension_list)
{
    if(glGetStringi_ptr || (glGetStringi_ptr =
       (const GLubyte*(GLApi*)(GLenum, GLuint))wglGetProcAddress("glGetStringi"))) {
        // try to fetch glGetStringi()

        r_s_extension_list.erase();

        int n_extension_num;
        glGetIntegerv(GL_NUM_EXTENSIONS, &n_extension_num);
        for(int i = 0; i < n_extension_num; ++ i) {
            const char *p_s_ext_name = (const char*)glGetStringi_ptr(GL_EXTENSIONS, i);
            if(!p_s_ext_name)
                continue; // ?
            try {
                if(i)
                    r_s_extension_list += " ";
                r_s_extension_list += p_s_ext_name;
            } catch(std::bad_alloc&) {
                return false;
            }
        }
        // new OpenGL 3 extensions scheme to avoid buffer overruns in old software

        return true;
    } else {
        const char *p_s_extensions = (const char*)glGetString(GL_EXTENSIONS);
        if(!p_s_extensions) {
            r_s_extension_list.erase();
            return false;
        }
        return stl_ut::AssignCStr(r_s_extension_list, p_s_extensions);
    }
}

std::string s_extensions_string; // storage for GL_EXTENSIONS string

const GLubyte *__gles20emupatch_glGetString(GLenum name)
{
    if(name == GL_EXTENSIONS) {
        CGLES20ExtensionHandler::Get_ExtensionString(s_extensions_string);
        return (const GLubyte*)s_extensions_string.c_str();
    } else
        return glGetString(name);
}

#if defined(_WIN32) || defined (_WIN64)

bool CGLES20ExtensionHandler::b_SupportedPlatformExtension(const char *p_s_extension_name)
{
#ifndef WGL_WGLEXT_PROTOTYPES
    PFNWGLGETEXTENSIONSSTRINGARBPROC wglGetExtensionsStringARB;
    if(!(wglGetExtensionsStringARB =
       (PFNWGLGETEXTENSIONSSTRINGARBPROC)wglGetProcAddress("wglGetExtensionsStringARB")))
        return false;
    // get wglGetExtensionsStringARB() if not linked statically
#endif // !WGL_WGLEXT_PROTOTYPES

    const char *p_s_extensions = (const char*)wglGetExtensionsStringARB(wglGetCurrentDC());
    // get WGL extensions string

    const char *p_s_search_str = p_s_extensions;
    do {
        if((p_s_search_str = strstr(p_s_search_str, p_s_extension_name))) {
            if(p_s_search_str > p_s_extensions && !isspace(*(p_s_search_str - 1)))
                continue;
            if(strlen(p_s_search_str + strlen(p_s_extension_name)) &&
               !isspace(*(p_s_search_str + strlen(p_s_extension_name))))
                continue;

            return true;
        }
    } while(p_s_search_str && strlen(++ p_s_search_str));
    // look for the extension

    return false;
}

#else // _WIN32 || _WIN64

bool CGLES20ExtensionHandler::b_SupportedPlatformExtension(const char *p_s_extension_name)
{
#ifdef GLX_VERSION_1_2
    const char *p_s_extensions;
    p_s_extensions = (const char*)glXGetClientString(glXGetCurrentDisplay(), GLX_EXTENSIONS); // @todo - debug this
    // glXGetClientString() is in GLX 1.1 and later (static linking)
    // glXGetCurrentDisplay() is in GLX 1.2 and later (static linking)
    // GLX_EXTENSIONS is in GLX 1.1
    // get GLX extensions string

    /*p_s_extensions = (const char*)glXQueryExtensionsString(glXGetCurrentDisplay(),
        DefaultScreen(glXGetCurrentDisplay()));*/
    // this could also work

    const char *p_s_search_str = p_s_extensions;
    do {
        if((p_s_search_str = strstr(p_s_search_str, p_s_extension_name))) {
            if(p_s_search_str > p_s_extensions && !isspace(*(p_s_search_str - 1)))
                continue;
            if(strlen(p_s_search_str + strlen(p_s_extension_name)) &&
               !isspace(*(p_s_search_str + strlen(p_s_extension_name))))
                continue;

            return true;
        }
    } while(p_s_search_str && strlen(++ p_s_search_str));
    // look for the extension

#else // GLX_VERSION_1_2
#error("GLX 1.2 is needed for glXGetClientString(GLX_EXTENSIONS) to work")
#endif // GLX_VERSION_1_2

    return false;
}

#endif // _WIN32 || _WIN64

int CGLES20ExtensionHandler::n_GetAllSupportedExtensionsFuncPointers()
{
#ifdef _DEBUG
    GLEH_OpenGL_ES_2_0 = false;
    GLEH_ARB_vertex_array_object = false;
    GLEH_OES_texture_3D = false;
    GLEH_OES_depth24 = false;
    GLEH_OES_depth32 = false;
    GLEH_OES_element_index_uint = false;
    GLEH_OES_mapbuffer = false;
    GLEH_OES_rgb8_rgba8 = false;
    GLEH_OES_stencil1 = false;
    GLEH_OES_stencil4 = false;
    GLEH_OES_stencil8 = false;
    GLEH_OES_depth_texture = false;
    GLEH_OES_packed_depth_stencil = false;
    GLEH_OES_get_program_binary = false;
    GLEH_EXT_texture_format_BGRA8888 = false;
    GLEH_NV_fence = false;
    GLEH_EXT_texture_lod_bias = false;
    GLEH_NV_draw_buffers = false;
    GLEH_NV_fbo_color_attachments = false;
    GLEH_NV_read_buffer = false;
    // this is just to check GLEH_ fral presence
#endif //_DEBUG

    return n_GetGLES20FuncPointers() +
           n_GetARBVertexArrayObjectFuncPointers() +
           n_GetOESTexture3DFuncPointers() +
           n_GetOESDepth24FuncPointers() +
           n_GetOESDepth32FuncPointers() +
           n_GetOESElementIndexUIntFuncPointers() +
           n_GetOESMapbufferFuncPointers() +
           n_GetOESRGB8RGBA8FuncPointers() +
           n_GetOESStencil1FuncPointers() +
           n_GetOESStencil4FuncPointers() +
           n_GetOESStencil8FuncPointers() +
           n_GetOESDepthTextureFuncPointers() +
           n_GetOESPackedDepthStencilFuncPointers() +
           n_GetOESGetProgramBinaryFuncPointers() +
           n_GetEXTTextureFormatBGRA8888FuncPointers() +
           n_GetNVFenceFuncPointers() +
           n_GetEXTTextureLODBiasFuncPointers() +
           n_GetNVDrawBuffersFuncPointers() +
           n_GetNVFBOColorAttachmentsFuncPointers() +
           n_GetNVReadBufferFuncPointers();
}

char *p_s_PreprocessShaderSource(const char *p_s_source, bool b_is_fragment_shader)
{
#ifdef _HAVE_OPENGLES2
	const char *p_s_source_code = p_s_source;
#else // _HAVE_OPENGLES2
	std::string s_source2;
	try {
		std::string s_temp_shader_source;
		do {
			int n_major, n_minor;
			const char *p_s_glsl_version = (const char*)glGetString(GL_SHADING_LANGUAGE_VERSION);
			if(!p_s_glsl_version || sscanf(p_s_glsl_version, "%d.%d", &n_major, &n_minor) != 2) {
				n_major = 0;
				n_minor = 0;
			}
			// try to obtain GLSL version

			if(CGLES20ExtensionHandler::b_Support_OpenGL(3, 3) && (n_major > 3 || (n_major == 3 && n_minor >= 30)))
				s_source2 = "#version 330\r\n"; // have to specify GLSL version on PC
			else if(n_major > 1 || (n_major == 1 && n_minor >= 30))
				s_source2 = "#version 130\r\n"; // this is fallback; some OpenGL 2.1 implementations with recent drivers do support this
			else {
				// no version specifier for very old implementations

				s_temp_shader_source = p_s_source;
				// copy shader source ...

				size_t n_pos = 0;
				while((n_pos = s_temp_shader_source.find("precision", n_pos)) != std::string::npos) {
					size_t b = n_pos, e = s_temp_shader_source.length();
					b += strlen("precision");
					// default precision specification statement

					if(b == e) {
						++ n_pos; // unknown syntax, just continue
						continue;
					}
					while(b < e && isspace(s_temp_shader_source[b]))
						++ b;
					// skip space

					if(b + 4 < e && !memcmp(s_temp_shader_source.data() + b, "lowp", 4 * sizeof(char)))
						b += 4;
					else if(b + 7 < e && !memcmp(s_temp_shader_source.data() + b, "mediump", 7 * sizeof(char)))
						b += 7;
					else if(b + 5 < e && !memcmp(s_temp_shader_source.data() + b, "highp", 5 * sizeof(char)))
						b += 5;
					else {
						++ n_pos; // unknown syntax, just continue
						continue;
					}
					// skip precission qualifier

					if(b == e) {
						++ n_pos; // unknown syntax, just continue
						continue;
					}
					while(b < e && isspace(s_temp_shader_source[b]))
						++ b;
					// skip space

					if(b == e) {
						++ n_pos; // unknown syntax, just continue
						continue;
					}
					while(b < e && isalnum(s_temp_shader_source[b]))
						++ b;
					// skip data type (eg. float, vec4, ...)

					while(b < e && isspace(s_temp_shader_source[b]))
						++ b;
					// skip space

					if(b == e || s_temp_shader_source[b] != ';') {
						++ n_pos; // unknown syntax, just continue
						continue;
					}
					++ b; // skip ';'
					// skip semicolon

					size_t n_begin = n_pos;
					size_t n_end = b;
					assert(n_end >= n_begin);
					if(n_end >= n_begin)
						s_temp_shader_source.erase(n_begin, n_end - n_begin);
					// erase the default precission specification
				}
				// this erases any default precision qualifiers there are

				// todo - it would also be nice to erase precision qualifiers from
				// the individual variables (eg. "uniform highp float time;")

				p_s_source = s_temp_shader_source.c_str();
				// just need to get rid of "precision highp float;" lines because these compilers do not understand those
				break;
			}
			// set shading language version

			if(b_is_fragment_shader) {
				// fragment
				s_source2 += "out vec4 _v_gl_frag_color_;\r\n";
				s_source2 += "#define gl_FragColor _v_gl_frag_color_\r\n";
				s_source2 += "#define varying in\r\n";
				s_source2 += "#line 0\r\n"; // fix line numbering ...
			} else {
				// vertex
				s_source2 += "#define attribute in\r\n";
				s_source2 += "#define varying out\r\n";
				s_source2 += "#line 0\r\n"; // fix line numbering ...
			};
			// add directives, if required
		} while(0);
		// add some directives for the OpenGL ES shaders to work on OpenGL hardware
		// todo - intercept glShaderSource and preprocess the source code in there

		s_source2 += p_s_source;
		// include the main shader source
	} catch(std::bad_alloc&) {
		return 0;
	}
	const char *p_s_source_code = s_source2.c_str();
#endif // _HAVE_OPENGLES2

#if defined(_MSC_VER) && !defined(__MWERKS__) && _MSC_VER >= 1400
	return _strdup(p_s_source_code);
#else // _MSC_VER && !__MWERKS__ && _MSC_VER >= 1400
	return strdup(p_s_source_code);
#endif // _MSC_VER && !__MWERKS__ && _MSC_VER >= 1400
}

char *p_s_PreprocessSource_VertexShader(const char *p_s_source)
{
    return p_s_PreprocessShaderSource(p_s_source, false);
}

char *p_s_PreprocessSource_FragmentShader(const char *p_s_source)
{
    return p_s_PreprocessShaderSource(p_s_source, true);
}

void __gles20emupatch_glShaderSource(GLuint shader, GLsizei count, const GLchar** string, const GLint* length)
{
    int n_type;
    glGetShaderiv(shader, GL_SHADER_TYPE, &n_type);

    char **p_source_list = new(std::nothrow) char*[count];
    // alloc list for shader sources (ignore alloc errors here)

    if(n_type == GL_VERTEX_SHADER) {
        for(int i = 0; i < count; ++ i)
            p_source_list[i] = p_s_PreprocessSource_VertexShader(string[i]);
        // preprocess the shaders
    } else if(n_type == GL_FRAGMENT_SHADER) {
        for(int i = 0; i < count; ++ i)
            p_source_list[i] = p_s_PreprocessSource_FragmentShader(string[i]);
        // preprocess the shaders
    } else {
		for(int i = 0; i < count; ++ i) {
#if defined(_MSC_VER) && !defined(__MWERKS__) && _MSC_VER >= 1400
			p_source_list[i] = _strdup(string[i]);
#else // _MSC_VER && !__MWERKS__ && _MSC_VER >= 1400
			p_source_list[i] = strdup(string[i]);
#endif // _MSC_VER && !__MWERKS__ && _MSC_VER >= 1400
		}
        // this is probably failure, someone didn't call glGenShaders(). let it fail ...
    }

    glShaderSource(shader, count, (const char**)p_source_list, length);
    // call GL

    for(int i = 0; i < count; ++ i)
        delete[] p_source_list[i];
    delete[] p_source_list;
    // cleanup
}

#define CGL4ExtensionHandler CGLES20ExtensionHandler
// fix code generated by GLEH
