
/*
 *  makes life a little bit easier when including gl.h by mistake
 */
#if defined(__gl2_h_) || defined(__GL_H__) || defined(GL_VERSION_1_1)
#error("warning: gl2.h must NOT be included in files using OpenGL30.h")
#endif

/*
 *  makes life a little bit easier when including glext.h by mistake
 */
#if defined(__glext_h_) || defined(GL_GLEXT_VERSION)
#error("glext.h must NOT be included in files using OpenGL30.h")
#endif

#define __gl2_h_
#define __gl_h_
#define __GL_H__
// does not include gl.h, defines GL up to 1.1 below, want to prevent gl.h
// header inclusion

#define __glext_h_
#define __GLEXT_H__
// glext.h is not included anymore, it does contain significant amount of
// deprecated functionality, want to prevent glext.h header inclusion

/**
 *  @def NULL_FUNCTION_POINTERS
 *  @brief if defined, initializes entry function pointers to null
 *
 *  In case NULL_FUNCTION_POINTERS is defined (default), all extension function pointers
 *      are initialized to 0, but user never needs to compare individual function values
 *      so it's posible to save some space in executable file by skipping the initialization.
 */
#define NULL_FUNCTION_POINTERS

/**
 *  @def FIX_GL_FUNCTION_LINKING_CONFLICTS
 *  @brief avoids function linking conflicts by renaming function pointers, and fixing
 *      names using #define
 *
 *  Some compilers may contain newer OpenGL libraries than this header requires
 *      (OpenGL 1.1, as supported by Microsoft Visual studio family), so names of function
 *      pointers, such as <tt>glBlendColor</tt> are present in static library, as well as in this
 *      file, causing link-time error. This is fixed by renaming the function, here
 *      <tt>glBlendColor_ptr</tt> would be used, and fixing it by macro, such as <tt>#define
 *      glBlendColor_ptr glBlendColor</tt>. The downside of this is, the Visual Studio edior
 *      will no longer show pop-up function parameters (intellisense), so this is best
 *      avoided if possible.
 */
#define FIX_GL_FUNCTION_LINKING_CONFLICTS

/**
 *  @def __GL_AUTOCHECK
 *  @brief automatically checks for OpenGL errors after every OpenGL function except glGetError()
 *  @note This is only available in debug mode unless FORCE_GL_CHECK is defined.
 */
#define __GL_AUTOCHECK

/**
 *  @def __GL_AUTOCHECK_USE_EXCEPTION
 *  @brief if defined, autocheck throws std::runtime_exception if an OpenGL error occurs
 */
//#define __GL_AUTOCHECK_USE_EXCEPTION

/**
 *  @def __GL_AUTOCHECK_USE_BREAKPOINT
 *  @brief if defined, autocheck calls user breakpoint if an OpenGL error occurs
 *  @note This is only available in windows / MSVC and in x86 mode (not x64).
 */
#define __GL_AUTOCHECK_USE_BREAKPOINT

/**
 *  @def __GL_AUTOCHECK_USE_ABORT
 *  @brief if defined, autocheck calls clib's abort() if an OpenGL error occurs
 */
//#define __GL_AUTOCHECK_USE_ABORT

/**
 *  @def __GL_AUTOCHECK_USE_EXIT
 *  @brief if defined, autocheck prints the error to stderr and calls
 *      exit(-1) if an OpenGL error occurs
 *  @note This is default behavior if nothing else is defined.
 */
//#define __GL_AUTOCHECK_USE_EXIT

/**
 *  @brief class for handling OpenGL extensions
 *
 *  Supplies means to get OpenGL function entry point, check extension support, or get
 *  OpenGL version. Functions with name n_Get*FuncPointers() returns number of function
 *  pointers that were not found (zero is success).
 */
class CGLES20ExtensionHandler {
public:
    /**
     *  @brief compares OpenGL version to reference
     *
     *  Compares supplied version to current OpenGL implementation version and returns
     *      true if current version is greater or equal (supplied version is supported).
     *
     *  @param[in] n_min_major is reference minimal major version number
     *  @param[in] n_min_minor is reference minimal minor version number
     *
     *  @return Returns true in case OpenGL version  is greater or equal to n_min_major.n_min_minor.
     */
    static bool b_Support_OpenGL(int n_min_major = 3, int n_min_minor = 0);

    /**
     *  @brief returns OpenGL version
     *
     *  Returns OpenGL version as (comparable) integer value.
     *
     *  @return Returns OpenGL version as major * 10 + minor (ie. for OpenGL 3.2 returns 32).
     */
    static int n_OpenGL_Version();

    /**
     *  @brief returns OpenGL version major
     *
     *  @return Returns OpenGL version as major (ie. for OpenGL 3.2 returns 3).
     */
    static int n_OpenGL_Version_Major();

    /**
     *  @brief returns OpenGL version minor
     *
     *  @return Returns OpenGL version as minor (ie. for OpenGL 3.2 returns 2).
     */
    static int n_OpenGL_Version_Minor();

    /**
     *  @brief returns OpenGL version as (major, minor) pair
     *
     *  @return Returns OpenGL version.
     */
    static std::pair<int, int> t_OpenGL_Version();

    /**
     *  @brief checks extension support
     *
     *  Checks support of a given OpenGL extension, returns true if extension is supported.
     *
     *  @param[in] p_s_extension_name is null-terminated string, containing name
     *      of OpenGL extension encoded in UTF-8
     *
     *  @return Returns true in case p_s_extension_name is supported, otherwise false.
     */
    static bool b_SupportedExtension(const char *p_s_extension_name);

    /**
     *  @brief checks (platform-specific) windowing system extension support
     *
     *  Checks support of a given WGL (GLX) extension, returns true if extension is supported.
     *
     *  @param[in] p_s_extension_name is null-terminated string, containing name
     *      of WGL (GLX) extension encoded in UTF-8
     *
     *  @return Returns true in case p_s_extension_name is supported, otherwise false.
     */
    static bool b_SupportedPlatformExtension(const char *p_s_extension_name);

    /**
     *  @brief gets extension string
     *
     *  OpenGL 3.0 doesn't have classic extension string, it returns names of individual
     *      extensions instead. This is to prevent buffer overruns, people always seem
     *      to be copying OpenGL extensions to fixed-size buffers, causing older software
     *      from times when extension string was shorter to fail on modern hardware.
     *      This functions returns "classic" OpenGL extension string in std::string,
     *      encoded in UTF-8, names of individual extensions are separated by space (' ')
     *      character.
     *
     *  @return Returns true on success, false on failure (not enough memory).
     */
    static bool Get_ExtensionString(std::string &r_s_extension_list);

    /**
     *  @brief gets entry points of all extensions
     *
     *  Gets pointers of function entry points for all supported extensions and OpenGL versions.
     *
     *  @return Returns number of missing functions (zero means success).
     *
     *  @note Note this is aimed at simple and test implementations without
     *      much need for error checking. it is always better to call individual
     *      functions, because in case this fails, there's no telling
     *      which extension exactly failed to initialize.
     */
    static int n_GetAllSupportedExtensionsFuncPointers();

    /**
     *  @brief gets entry points of all OpenGL ES 2.0 core functions
     *
     *  @return Returns number of missing functions (zero means success).
     */
    static int n_GetGLES20FuncPointers();
